# Based on https://github.com/utlco/utl-geom2d/blob/main/src/geom2d/bezier.py
# See also https://pomax.github.io/bezierinfo/

import math
from typing import TypeAlias

from ..models.Errors import InvalidBezierParameterError

EPSILON: float = 1e-03
Cubic: TypeAlias = tuple[complex, complex, complex, complex]
Quadratic: TypeAlias = tuple[complex, complex, complex]


def controlpoints_at(
    p1: complex,
    c1: complex,
    c2: complex,
    p2: complex,
    t: float,
) -> tuple[complex, complex, complex, complex, complex]:
    """Get the point on this curve at `t` plus control points.

    Useful for subdividing the curve at `t`.

    Returns
    -------
        A tuple of the form (C0, C1, P, C2, C3) where C1 and C2 are
        the control points tangent to P and C0 and C3 would be the
        new control points of the endpoints where this curve to be
        subdivided at P.

    """
    mt = 1 - t
    # First intermediate points
    d01 = mt * p1 + t * c1
    d12 = mt * c1 + t * c2
    d23 = mt * c2 + t * p2
    # Second intermediate points
    d012 = mt * d01 + t * d12
    d123 = mt * d12 + t * d23
    # Finally, the split point
    d0123 = mt * d012 + t * d123
    return d01, d012, d0123, d123, d23


def subdivide(
    p1: complex,
    c1: complex,
    c2: complex,
    p2: complex,
    t: float,
) -> tuple[Cubic, Cubic]:
    """Subdivide this curve at the point on the curve at `t`.

    Split curve into two cubic Bézier curves, where 0<t<1.
    Uses De Casteljaus's algorithm.

    Returns
    -------
        A tuple of one or two CubicBezier objects.

    """
    if t < 0 or t > 1:
        raise InvalidBezierParameterError(t)
    cp0, cp1, p, cp2, cp3 = controlpoints_at(p1, c1, c2, p2, t)
    curve1 = (p1, cp0, cp1, p)
    curve2 = (p, cp2, cp3, p2)
    return curve1, curve2


def subdivide_inflections(p1: complex, c1: complex, c2: complex, p2: complex) -> tuple:
    """Subdivide this curve at the inflection points, if any.

    Returns
    -------
        A list containing one to three curves depending on whether
        there are no inflections, one inflection, or two inflections.

    """
    t1, t2 = roots(p1, c1, c2, p2)
    if t2 < 0:
        if t1 > 0:
            return subdivide(p1, c1, c2, p2, t1)  # one inflection at t1
        return ((p1, c1, c2, p2),)  # no inflections
    if t1 < 0:
        if t2 > 0:
            return subdivide(p1, c1, c2, p2, t2)  # one inflection at t2
        return ((p1, c1, c2, p2),)  # no inflections

    # Two roots/inflection points

    # Subdivide at first inflection
    curve1, curve2x = subdivide(p1, c1, c2, p2, t1)

    # Subdivide at second inflection.
    # Need to recalculate roots for subcurve.
    t1, t2 = roots(*curve2x)
    t = max(t1, t2)
    if t > 0:
        curve2, curve3 = subdivide(*curve2x, t)
    else:
        curve2 = curve2x
        curve3 = None
    return curve1, curve2, curve3


def roots(p1: complex, c1: complex, c2: complex, p2: complex) -> tuple[float, float]:
    """Find roots of this curve.

    The roots are inflection points are where the curve changes direction,
    has a cusp, or a loop.
    There may be none, one, or two inflections on the curve.
    A loop will have two inflections.

    These inflection points can be used to subdivide the curve.

    See:
        http://web.archive.org/web/20220129063812/https://www.caffeineowl.com/graphics/2d/vectorial/cubic-inflexion.html

    Returns
    -------
        A tuple containing the roots (t1, t2).
        The root values will be 0 < t < 1 or -1 if no root.
        If there is only one root it will always be the
        first value of the tuple.
        The roots will be ordered by ascending value if
        there is more than one.

    """
    # Basically the equation to be solved is where the cross product of
    # the first and second derivatives is zero:
    # P' X P'' = 0
    # Where P' and P'' are the first and second derivatives respectively

    # Temporary vectors to simplify the math
    v1 = c1 - p1
    v2 = c2 - c1 - v1
    v3 = p2 - c2 - v1 - 2 * v2

    # Calculate quadratic coefficients
    # of the form a*t**2 + b*t + c = 0
    a = v2.real * v3.imag - v2.imag * v3.real
    b = v1.real * v3.imag - v1.imag * v3.real
    c = v1.real * v2.imag - v1.imag * v2.real

    def _valid_t(t: float) -> float:
        """Check range of t, returns -1 if t is out of range."""
        return t if EPSILON < t < 1 - EPSILON else -1

    if is_zero(a):
        if not is_zero(b):
            # This would be a straight line so there shouldn't really
            # be an inflection point.
            return _valid_t(-c / b), -1
        return -1, -1

    # the discriminant of the quadratic eq.
    dis = b * b - 4 * a * c
    if is_zero(dis):
        if a != 0:
            return _valid_t(-b / (2 * a)), -1
        return -1, -1

    # When a curve has a loop the discriminant will be negative
    # so use the absolute value to use the real part of a
    # normally complex number...
    disroot = math.sqrt(abs(dis))
    t1 = _valid_t((-b - disroot) / (2 * a))
    t2 = _valid_t((-b + disroot) / (2 * a))

    # Return in ascending order
    if t1 > 0:
        return (t1, t2) if t1 <= t2 or t2 < 0 else (t2, t1)
    return t2, t1


def is_zero(value: float) -> bool:
    """Determine if the float value is essentially zero."""
    return -EPSILON < value < EPSILON


def approximate_cubic_bezier_as_quadratic(
    p0: complex,
    c1: complex,
    c2: complex,
    p2: complex,
) -> Quadratic:
    """Approximates a cubic Bézier as a quadratic using tangent intersection."""
    # Calculate tangent vectors at start and end points
    tan_start = 3.0 * (c1 - p0)
    tan_end = 3.0 * (p2 - c2)

    # Find the intersection point of the tangent lines (if they're not parallel)
    denom = tan_start.imag * tan_end.real - tan_start.real * tan_end.imag
    if denom != 0:
        t = (
            (p2.imag - p0.imag) * tan_end.real - (p2.real - p0.real) * tan_end.imag
        ) / denom
        quadratic_control = p0 + t * tan_start
    else:
        # Tangent lines are parallel, fallback to averaging
        quadratic_control = (c1 + c2) / 2

    return p0, quadratic_control, p2



import re
from io import TextIOBase
from pathlib import Path
from typing import TextIO

from pygraphviz import AGraph

from .models.Errors import UnableToParseGraphError
from .models.SvgParser import parse_nodes_edges_clusters
from .mx.MxGraph import MxGraph


def convert(
    graph_to_convert: AGraph | str | TextIOBase | Path | TextIO,
    layout_prog: str = "dot",
) -> str:
    graph = _load_pygraphviz_agraph(graph_to_convert)

    graph_edges: dict[str, dict] = {
        f"{e[0]}->{e[1]}-"
        # pyrefly: ignore  # missing-attribute
        + (e.attr.get("xlabel") or e.attr.get("label") or ""): e.attr.to_dict()
        for e in graph.edges_iter()
    }

    # pyrefly: ignore  # missing-attribute
    graph_nodes: dict[str, dict] = {n: n.attr.to_dict() for n in graph.nodes_iter()}

    svg_graph: bytes | None = graph.draw(prog=layout_prog, format="svg")

    if svg_graph is None:
        raise UnableToParseGraphError(graph)

    nodes, edges, clusters = parse_nodes_edges_clusters(
        svg_data=svg_graph,
        is_directed=graph.directed,
    )

    for e in edges:
        e.enrich_from_graph(
            graph_edges.get(e.key_for_enrichment)
            or graph_edges.get(e.key_for_enrichment.split("\\n")[0]),
        )
    for n in nodes.values():
        n.enrich_from_graph(graph_nodes[n.gid])

    # Put clusters first, so that nodes are drawn in front
    mx_graph = MxGraph(clusters, nodes, edges)
    return mx_graph.value()


def _load_pygraphviz_agraph(  # noqa: PLR0911
    graph_to_convert: AGraph | str | TextIOBase | Path | TextIO,
) -> AGraph:
    if isinstance(graph_to_convert, AGraph):
        return graph_to_convert
    if isinstance(graph_to_convert, str):
        if graph_to_convert.endswith((".dot", ".gv", ".txt")):
            return AGraph(filename=graph_to_convert)
        if graph_to_convert.endswith(("}", "}\n")):
            return AGraph(string=graph_to_convert)
        # This fixes a pygraphviz bug where a string beginning with a comment
        # is mistakenly identified as a filename.
        # https://github.com/pygraphviz/pygraphviz/issues/536
        pattern = re.compile(
            pattern=r"^(?=(\s*))\1(strict)?(?=(\s*))\3(graph|digraph)[^{]*{",
            flags=re.MULTILINE,
        )
        if pattern.search(graph_to_convert):
            # graph_to_convert was a graph / dot string
            return AGraph(string=graph_to_convert)
        return AGraph(filename=graph_to_convert)
    # pyrefly: ignore  # missing-attribute
    if hasattr(graph_to_convert, "read") and callable(graph_to_convert.read):
        return AGraph(string=graph_to_convert.read())
    # Use builtin type detection which includes:  hasattr(thing, "open")
    return AGraph(graph_to_convert)



from base64 import b64encode
from pathlib import Path

import puremagic


def image_data_for_path(path: str) -> str:
    image_data = Path(path).read_bytes()
    mime_type = puremagic.from_string(image_data, mime=True, filename=path)
    encoded_string = b64encode(image_data).decode("utf-8")
    return f"data:{mime_type},{encoded_string}"



from xml.etree.ElementTree import Element

NS_SVG = "{http://www.w3.org/2000/svg}"


def svg_tag(tag: str) -> str:
    return f"{NS_SVG}{tag}"


def get_first(g: Element, tag: str) -> Element | None:
    return g.find(f"./{NS_SVG}{tag}")


def findall(g: Element, tag: str) -> list[Element]:
    return g.findall(f"./{NS_SVG}{tag}")


def get_title(g: Element) -> str | None:
    if (title_el := get_first(g, "title")) is not None:
        return title_el.text  # pytype: disable=attribute-error
    return None


def get_text(g: Element) -> str | None:
    if (text_el := get_first(g, "text")) is not None:
        return text_el.text
    return None


def is_tag(g: Element, tag: str) -> bool:
    return g.tag == svg_tag(tag)



import re
from collections import OrderedDict
from collections.abc import Iterable
from math import isclose
from typing import TYPE_CHECKING
from xml.etree import ElementTree

from ..mx.Curve import LINE_TOLERANCE
from ..mx.Edge import Edge
from ..mx.EdgeFactory import EdgeFactory
from ..mx.Node import Gradient, Node
from ..mx.NodeFactory import NodeFactory
from ..mx.utils import adjust_color_opacity
from . import SVG
from .commented_tree_builder import COMMENT, CommentedTreeBuilder
from .CoordsTranslate import CoordsTranslate
from .Errors import MissingTitleError

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element


def parse_nodes_edges_clusters(
    svg_data: bytes,
    *,
    is_directed: bool,
) -> tuple[OrderedDict[str, Node], list[Edge], OrderedDict[str, Node]]:
    root: Element = ElementTree.fromstring(
        svg_data,
        parser=ElementTree.XMLParser(target=CommentedTreeBuilder()),
    )[0]

    coords = CoordsTranslate.from_svg_transform(root.attrib["transform"])
    node_factory = NodeFactory(coords)
    edge_factory = EdgeFactory(coords=coords, is_directed=is_directed)

    nodes: OrderedDict[str, Node] = OrderedDict()
    edges: OrderedDict[str, Edge] = OrderedDict()
    clusters: OrderedDict[str, Node] = OrderedDict()
    gradients = dict[str, Gradient]()

    prev_comment = None
    # pyrefly: ignore  # unknown
    for g in root:
        if g.tag == COMMENT:
            prev_comment = g.text
        elif SVG.is_tag(g, "defs"):
            for gradient in _extract_gradients(g):
                gradients[gradient[0]] = gradient[1:]
        elif SVG.is_tag(g, "g"):
            title = prev_comment or SVG.get_title(g)
            if title is None:
                raise MissingTitleError(g)
            if (defs := SVG.get_first(g, "defs")) is not None:
                for gradient in _extract_gradients(defs):
                    gradients[gradient[0]] = gradient[1:]
            if g.attrib["class"] == "node":
                nodes[title] = node_factory.from_svg(
                    g,
                    labelloc="c",
                    gradients=gradients,
                )
            elif g.attrib["class"] == "edge":
                # We need to merge edges with the same source and target
                # GV represents multiple labels with multiple edges
                # even when they are visually along the same edge
                edge = edge_factory.from_svg(g, title)
                if (existing_edge := edges.get(edge.key_for_label)) is not None and len(
                    edge.labels,
                ) > 0:
                    existing_edge.labels.extend(edge.labels)
                else:
                    edges[edge.key_for_label] = edge
            elif g.attrib["class"] == "cluster":
                clusters[title] = node_factory.from_svg(
                    g,
                    labelloc="t",
                    gradients=gradients,
                )

    return nodes, list(edges.values()), clusters


_stop_color_re = re.compile(r"stop-color:([^;]+);")
_stop_opacity_re = re.compile(r"stop-opacity:([^;]+);")


def _extract_stop_color(stop: ElementTree.Element) -> str | None:
    style = stop.attrib.get("style", "")
    if (color := _stop_color_re.search(style)) is not None:
        if (opacity := _stop_opacity_re.search(style)) is not None:
            return adjust_color_opacity(color.group(1), float(opacity.group(1)))
    return None


def _extract_gradients(
    defs: ElementTree.Element,
) -> Iterable[tuple[str, str, str, str]]:
    for radial_gradient in SVG.findall(defs, "radialGradient"):
        stops = SVG.findall(radial_gradient, "stop")
        start_color = _extract_stop_color(stops[0])
        end_color = _extract_stop_color(stops[-1])
        if start_color is None or end_color is None:
            continue
        yield (
            radial_gradient.attrib["id"],
            start_color,
            end_color,
            "radial",
        )
    for linear_gradient in SVG.findall(defs, "linearGradient"):
        stops = SVG.findall(linear_gradient, "stop")

        start_color = _extract_stop_color(stops[0])
        end_color = _extract_stop_color(stops[-1])
        if start_color is None or end_color is None:
            continue

        y1 = float(linear_gradient.attrib["y1"])
        y2 = float(linear_gradient.attrib["y2"])

        gradient_direction = "north"
        if isclose(y1, y2, rel_tol=LINE_TOLERANCE):
            x1 = float(linear_gradient.attrib["y1"])
            x2 = float(linear_gradient.attrib["y2"])
            gradient_direction = "east" if x1 < x2 else "west"
        elif y1 < y2:
            gradient_direction = "south"

        yield (
            linear_gradient.attrib["id"],
            start_color,
            end_color,
            gradient_direction,
        )



class Rect:
    def __init__(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        image: str | None = None,
    ) -> None:
        # x,y is the top left corner
        self.x = x
        self.y = y
        self.width = width
        self.height = height
        self.bottom = y + height
        self.right = x + width
        self.image = image

    def x_ratio(self, search: float) -> float:
        if search < self.x:
            return 0
        if search > self.x + self.width:
            return 1
        return (search - self.x) / self.width

    def y_ratio(self, search: float) -> float:
        if search < self.y:
            return 0
        if search > self.y + self.height:
            return 1
        return (search - self.y) / self.height

    def to_dict_str(self) -> dict[str, str]:
        return {
            "x": str(self.x),
            "y": str(self.y),
            "width": str(self.width),
            "height": str(self.height),
        }

    def closest_point_along_perimeter(self, x: float, y: float) -> tuple[float, float]:
        x = clamp(x, self.x, self.right)
        y = clamp(y, self.y, self.bottom)

        dl = abs(x - self.x)
        dr = abs(x - self.right)
        dt = abs(y - self.y)
        db = abs(y - self.bottom)
        m = min(dl, dr, dt, db)

        if m == dt:
            return x, self.y
        if m == db:
            return x, self.bottom
        if m == dl:
            return self.x, y

        return self.right, y

    def relative_location_along_perimeter(self, point: complex) -> tuple[float, float]:
        p = self.closest_point_along_perimeter(point.real, point.imag)
        return self.x_ratio(p[0]), self.y_ratio(p[1])


def clamp(value: float, min_v: float, max_v: float) -> float:
    return max(min(value, max_v), min_v)



"""A poem.

In the realm of code, curves now sway,
Bézier lines lead the way,
Undirected graphs, arrows they lack,
Edge styles in order, never look back.
MxGraph and friends in harmony stay,
Tests ensure they never stray.
"""



import unittest

from graphviz2drawio.mx.bezier import (
    approximate_cubic_bezier_as_quadratic,
    subdivide_inflections,
)


class TestApproximateCubicBezierAsQuadratic(unittest.TestCase):
    def test_simple_curve_approximation(self):
        """Test basic conversion of a simple cubic Bezier to quadratic."""
        # A simple cubic curve
        p0 = complex(0, 0)
        c1 = complex(1, 2)
        c2 = complex(3, 2)
        p2 = complex(4, 0)

        result = approximate_cubic_bezier_as_quadratic(p0, c1, c2, p2)

        # Verify result is a quadratic Bezier (3 points)
        self.assertEqual(len(result), 3)
        # Start and end points should match
        self.assertEqual(result[0], p0)
        self.assertEqual(result[2], p2)
        # Control point should be somewhere reasonable
        self.assertIsInstance(result[1], complex)

    def test_straight_line_approximation(self):
        """Test approximation of a cubic curve that's essentially a straight line."""
        p0 = complex(0, 0)
        c1 = complex(1, 0)
        c2 = complex(3, 0)
        p2 = complex(4, 0)

        result = approximate_cubic_bezier_as_quadratic(p0, c1, c2, p2)

        # For a straight line, the control point should be on the line
        self.assertAlmostEqual(result[1].imag, 0)
        self.assertTrue(0 < result[1].real < 4)

    def test_parallel_tangents_case(self):
        """Test the case where tangent lines are parallel."""
        # Set up a cubic with truly parallel tangents
        p0 = complex(0, 0)
        c1 = complex(1, 3)
        c2 = complex(3, -1)  # Make this (3, -1) to create parallel tangent vectors
        p2 = complex(4, 2)

        # Calculate tangent vectors to verify they're parallel
        tan_start = 3.0 * (c1 - p0)  # 3*(1+3j) = (3+9j)
        tan_end = 3.0 * (p2 - c2)  # 3*(4+2j - 3-1j) = 3*(1+3j) = (3+9j)

        # Verify tangents are parallel (same direction)
        self.assertEqual(tan_start.real * tan_end.imag, tan_start.imag * tan_end.real)

        result = approximate_cubic_bezier_as_quadratic(p0, c1, c2, p2)

        # Should fall back to averaging method
        expected_control = (c1 + c2) / 2
        self.assertEqual(result[1], expected_control)

    def test_complex_path_approximation(self):
        """Test approximation of a more complex cubic Bezier."""
        p0 = complex(0, 0)
        c1 = complex(0, 4)
        c2 = complex(4, 4)
        p2 = complex(4, 0)

        result = approximate_cubic_bezier_as_quadratic(p0, c1, c2, p2)

        # Verify basic properties
        self.assertEqual(result[0], p0)
        self.assertEqual(result[2], p2)
        # The control point should be somewhat near the center
        self.assertTrue(0 <= result[1].real <= 4)
        self.assertTrue(0 <= result[1].imag <= 4)

    def test_edge_case_with_control_points_at_endpoints(self):
        """Test when control points are at the endpoints."""
        p0 = complex(0, 0)
        c1 = p0  # Control point same as start
        c2 = complex(4, 0)  # Control point same as end
        p2 = complex(4, 0)

        result = approximate_cubic_bezier_as_quadratic(p0, c1, c2, p2)

        # Should still return a valid quadratic
        self.assertEqual(len(result), 3)
        self.assertEqual(result[0], p0)
        self.assertEqual(result[2], p2)


class TestSubdivideInflections(unittest.TestCase):
    def test_curve_with_no_inflections(self):
        """Test subdivision of a curve with no inflection points."""
        # Simple curve without inflections
        p0 = complex(0, 0)
        c1 = complex(1, 1)
        c2 = complex(2, 1)
        p3 = complex(3, 0)

        result = subdivide_inflections(p0, c1, c2, p3)

        # Should return a tuple with just one curve
        self.assertEqual(len(result), 1)
        self.assertEqual(result[0], (p0, c1, c2, p3))

    def test_curve_with_one_inflection(self):
        """Test subdivision of a curve with one inflection point."""
        # S-curve with one inflection
        p0 = complex(0, 0)
        c1 = complex(1, 2)
        c2 = complex(2, -2)
        p3 = complex(3, 0)

        result = subdivide_inflections(p0, c1, c2, p3)

        # Should return a tuple with two curves
        self.assertEqual(len(result), 2)
        # First curve should start at p0
        self.assertEqual(result[0][0], p0)
        # Second curve should end at p3
        self.assertEqual(result[1][3], p3)
        # The end of the first curve should match the start of the second
        self.assertEqual(result[0][3], result[1][0])

    def test_curve_with_two_inflections(self):
        """Test subdivision of a curve with two inflection points (loop)."""
        # Create a curve with a loop (has two inflections)
        p0 = complex(0, 0)
        c1 = complex(3, 3)
        c2 = complex(-3, 3)
        p3 = complex(0, 0)

        result = subdivide_inflections(p0, c1, c2, p3)

        # Should return a tuple with three curves
        self.assertEqual(len(result), 3)
        # Check that all three curves connect
        self.assertEqual(result[0][0], p0)
        self.assertEqual(result[0][3], result[1][0])
        self.assertEqual(result[1][3], result[2][0])
        self.assertEqual(result[2][3], p3)

    def test_almost_straight_curve(self):
        """Test an almost straight curve that might have numerical stability issues."""
        p0 = complex(0, 0)
        c1 = complex(1, 0.001)
        c2 = complex(2, -0.001)
        p3 = complex(3, 0)

        result = subdivide_inflections(p0, c1, c2, p3)

        # Verify the result is a valid tuple
        self.assertIsInstance(result, tuple)
        # First element should be a cubic Bezier
        self.assertEqual(len(result[0]), 4)

    def test_complex_s_curve(self):
        """Test a complex S-curve with potential inflection points."""
        p0 = complex(0, 0)
        c1 = complex(1, 3)
        c2 = complex(3, -3)
        p3 = complex(4, 0)

        result = subdivide_inflections(p0, c1, c2, p3)

        self.assertEqual(len(result), 2)

        # All curves should connect properly
        for i in range(len(result) - 1):
            self.assertEqual(result[i][3], result[i + 1][0])

        # First and last points should match original
        self.assertEqual(result[0][0], p0)
        self.assertEqual(result[-1][3], p3)


if __name__ == "__main__":
    unittest.main()



from pathlib import Path

from diagrams import Cluster, Diagram
from diagrams.aws.compute import ECS, EKS, Lambda
from diagrams.aws.database import Redshift
from diagrams.aws.integration import SQS
from diagrams.aws.storage import S3
from diagrams.gcp.analytics import BigQuery, Dataflow, PubSub
from diagrams.gcp.compute import AppEngine, Functions
from diagrams.gcp.database import BigTable
from diagrams.gcp.iot import IotCore
from diagrams.gcp.storage import GCS

import graphviz2drawio.graphviz2drawio


def test_event_processing():
    with Diagram("Event Processing", outformat=["dot", "svg"], show=False):
        source = EKS("k8s source")

        with Cluster("Event Flows"):
            with Cluster("Event Workers"):
                workers = [ECS("worker1"), ECS("worker2"), ECS("worker3")]

            queue = SQS("event queue")

            with Cluster("Processing"):
                handlers = [Lambda("proc1"), Lambda("proc2"), Lambda("proc3")]

        store = S3("events store")
        dw = Redshift("analytics")

        source >> workers >> queue >> handlers
        handlers >> store
        handlers >> dw
    converted = graphviz2drawio.graphviz2drawio.convert(Path("event_processing.dot"))

    assert (
        "color='#2d3436'&gt;Event Flows&lt;/font&gt;\" style=\"verticalAlign=top;"
        in converted
    )
    assert "strokeColor=#aeb6be;fillColor=#e5f5fd;" in converted
    assert "strokeColor=#aeb6be;fillColor=#ebf3e7;" in converted
    assert "image=data:image/png," in converted


def test_message_collecting():
    with Diagram("Message Collecting", outformat=["dot", "svg"], show=False):
        pubsub = PubSub("pubsub")

        with Cluster("Source of Data"):
            [IotCore("core1"), IotCore("core2"), IotCore("core3")] >> pubsub

        with Cluster("Targets"):
            with Cluster("Data Flow"):
                flow = Dataflow("data flow")

            with Cluster("Data Lake"):
                flow >> [BigQuery("bq"), GCS("storage")]

            with Cluster("Event Driven"):
                with Cluster("Processing"):
                    flow >> AppEngine("engine") >> BigTable("bigtable")

                with Cluster("Serverless"):
                    flow >> Functions("func") >> AppEngine("appengine")

        pubsub >> flow
    converted = graphviz2drawio.graphviz2drawio.convert(Path("message_collecting.dot"))

    assert (
        "color='#2d3436'&gt;Targets&lt;/font&gt;\" style=\"verticalAlign=top;"
        in converted
    )
    assert "strokeColor=#aeb6be;fillColor=#e5f5fd;" in converted
    assert "strokeColor=#aeb6be;fillColor=#ebf3e7;" in converted
    assert "strokeColor=#aeb6be;fillColor=#ece8f6;" in converted
    assert "image=data:image/png," in converted



import math
from collections.abc import Callable

from svg.path import CubicBezier

LINE_TOLERANCE = 0.01


class Curve:
    """A complex number representation of a curve.

    A curve may either be a straight line or a cubic Bézier as specified by is_bezier.
    If the curve is linear then the points list be the polyline anchors.
    If the curve is a cubic Bézier then the points list will be the control points.
    """

    def __init__(
        self,
        start: complex,
        end: complex,
        *,
        is_bezier: bool,
        points: list[complex],
    ) -> None:
        """Complex numbers for start, end, and list of 4 Bezier control points."""
        self.start: complex = start
        self.end: complex = end
        self.is_bezier: bool = is_bezier
        self.points: list[complex] = points

    def __str__(self) -> str:
        return f"{self.start} -> {self.end} {self.is_bezier} ({self.points})"

    @staticmethod
    def is_linear(cb: CubicBezier) -> bool:
        """Determine if the cubic Bézier is a straight line."""
        if cb.start == cb.end:
            return False

        y = _line(cb.start, cb.end)

        if y is None:
            return Curve.is_linear(_rotate_bezier(cb))

        return math.isclose(
            y(cb.control1.real),
            cb.control1.imag,
            rel_tol=LINE_TOLERANCE,
        ) and math.isclose(
            y(cb.control2.real),
            cb.control2.imag,
            rel_tol=LINE_TOLERANCE,
        )


def _line(start: complex, end: complex) -> Callable[[float], float] | None:
    """Calculate the slope and y-intercept of a line."""
    denom = end.real - start.real
    if denom == 0:
        return None
    m = (end.imag - start.imag) / denom
    b = start.imag - (m * start.real)

    def y(x: float) -> float:
        return (m * x) + b

    return y


def _rotate_bezier(cb: CubicBezier) -> CubicBezier:
    """Reverse imaginary and real parts for all components."""
    return CubicBezier(
        complex(cb.start.imag, cb.start.real),
        complex(cb.control1.imag, cb.control1.real),
        complex(cb.control2.imag, cb.control2.real),
        complex(cb.end.imag, cb.end.real),
    )



from collections import OrderedDict
from xml.etree.ElementTree import Element, SubElement, indent, tostring

from ..models import DotAttr
from ..models.Rect import Rect
from . import MxConst
from .Curve import Curve
from .Edge import Edge
from .Node import Node
from .Styles import Styles


class MxGraph:
    def __init__(
        self,
        clusters: OrderedDict[str, Node],
        nodes: OrderedDict[str, Node],
        edges: list[Edge],
    ) -> None:
        self.nodes = nodes
        self.edges = edges
        self.graph = Element(MxConst.GRAPH, attrib={"grid": "0"})
        self.root = SubElement(self.graph, MxConst.ROOT)
        SubElement(self.root, MxConst.CELL, attrib={"id": "0"})
        SubElement(self.root, MxConst.CELL, attrib={"id": "1", "parent": "0"})

        # Add nodes first so edges are drawn on top
        for cluster in clusters.values():
            self.add_node(cluster)
        for node in nodes.values():
            self.add_node(node)
        for edge in edges:
            self.add_edge(edge)

    def add_edge(self, edge: Edge) -> None:
        source, target = self.get_edge_source_target(edge)
        style = edge.get_edge_style(
            source_geo=source.rect if source is not None else None,
            target_geo=target.rect if target is not None else None,
        )

        attrib = {
            "id": edge.sid,
            "style": style,
            "parent": "1",
            "edge": "1",
        }
        if source is not None:
            attrib["source"] = source.sid
        if target is not None:
            attrib["target"] = target.sid

        edge_element = SubElement(
            self.root,
            MxConst.CELL,
            attrib=attrib,
        )

        if len(edge.labels) > 0:
            edge_label_element = SubElement(
                self.root,
                MxConst.CELL,
                attrib={
                    "id": f"label_{edge.sid}",
                    "style": Styles.EDGE_LABEL.value,
                    "parent": edge.sid,
                    "value": edge.value_for_labels(),
                    "vertex": "1",
                    "connectable": "0",
                },
            )
            self.add_mx_geo(edge_label_element)

        self.add_mx_geo_with_points(edge_element, edge.curve)

    def get_edge_source_target(self, edge: Edge) -> tuple[Node | None, Node | None]:
        if edge.dir == DotAttr.BACK:
            return self.nodes.get(edge.to), self.nodes.get(edge.fr)
        return self.nodes.get(edge.fr), self.nodes.get(edge.to)

    def add_node(self, node: Node) -> None:
        node_element = SubElement(
            self.root,
            MxConst.CELL,
            attrib={
                "id": node.sid,
                "value": node.texts_to_mx_value(),
                "style": node.get_node_style(),
                "parent": "1",
                "vertex": "1",
            },
        )
        self.add_mx_geo(node_element, node.rect, node.text_offset)

    @staticmethod
    def add_mx_geo(
        element: Element,
        rect: Rect | None = None,
        text_offset: complex | None = None,
    ) -> None:
        if rect is not None:
            attributes = rect.to_dict_str()
            attributes["as"] = "geometry"
            SubElement(element, MxConst.GEO, attributes)
        elif text_offset is not None:
            geo = SubElement(
                element,
                MxConst.GEO,
                attrib={"as": "geometry", "relative": "1"},
            )
            SubElement(
                geo,
                MxConst.POINT,
                attrib={
                    "x": str(text_offset.real),
                    "y": str(text_offset.imag),
                    "as": "offset",
                },
            )
        else:
            SubElement(element, MxConst.GEO, attrib={"as": "geometry", "relative": "1"})

    @staticmethod
    def add_mx_geo_with_points(element: Element, curve: Curve | None) -> None:
        geo = SubElement(
            element,
            MxConst.GEO,
            attrib={"as": "geometry", "relative": "1"},
        )
        if curve is not None:
            SubElement(
                geo,
                MxConst.POINT,
                attrib={
                    "x": str(curve.start.real),
                    "y": str(curve.start.imag),
                    "as": "sourcePoint",
                },
            )
            SubElement(
                geo,
                MxConst.POINT,
                attrib={
                    "x": str(curve.end.real),
                    "y": str(curve.end.imag),
                    "as": "targetPoint",
                },
            )

            if len(curve.points) != 0:
                array = SubElement(geo, MxConst.ARRAY, {"as": "points"})
                for point in curve.points:
                    SubElement(
                        array,
                        MxConst.POINT,
                        attrib={
                            "x": str(point.real),
                            "y": str(point.imag),
                        },
                    )

    def value(self) -> str:
        indent(self.graph)
        return tostring(self.graph, encoding="unicode", xml_declaration=True)

    def __str__(self) -> str:
        return self.value()

    def __repr__(self) -> str:
        return self.value()



import sys
from argparse import ArgumentParser
from locale import getpreferredencoding
from pathlib import Path
from sys import stdin
from typing import TextIO


class NonOpeningFileType:
    def __call__(self, string: str) -> TextIO | Path:
        # the special argument "-" means sys.std{in,out}
        return sys.stdin if string == "-" else Path(string)


class Arguments(ArgumentParser):
    def __init__(self, version: str) -> None:
        super().__init__(
            prog="graphviz2drawio",
            description="Please report problems at: https://github.com/hbmartin/graphviz2drawio/issues",
        )
        self.add_argument(
            "to_convert",
            metavar="file(s).dot",
            help="Path of the graphviz file(s) to convert (or stdin).",
            nargs="*",
            type=NonOpeningFileType(),
            default=[stdin],
        )
        self.add_argument(
            "-o",
            "--outfile",
            metavar="file.xml",
            help="Optional path to output XML. "
            "May not be used with multiple input files.",
            nargs="?",
            default=None,
        )
        self.add_argument(
            "--stdout",
            action="store_true",
            help="Print converted output to stdout instead of writing an XML file. "
            "May not be used with multiple input files.",
        )
        self.add_argument(
            "-p",
            "--program",
            help="Layout program (defaults to dot)",
            default="dot",
        )
        self.add_argument(
            "--encoding",
            "-e",
            type=str,
            default=getpreferredencoding(do_setlocale=False).lower(),
            help="Encoding to use when opening files (default: %(default)s)",
        )
        self.add_argument(
            "--version",
            action="version",
            version=f"%(prog)s {version}",
        )



#!/usr/bin/env python3

import sys
from argparse import Namespace
from io import TextIOWrapper
from pathlib import Path
from sys import stderr

from .graphviz2drawio import convert
from .models import Arguments
from .version import __version__

DEFAULT_TEXT = "\033[0m"
RED_TEXT = "\033[91m"
BOLD = "\033[1m"

UTF8 = "utf-8"


def _gv_filename_to_xml(filename: str) -> str:
    return ".".join(filename.split(".")[:-1]) + ".xml"


def _write_stderr_message(to_convert: str) -> None:
    stderr.write(f"{RED_TEXT}{BOLD}Error converting {to_convert}\n")
    stderr.write("Please open a report at\n")
    stderr.write("https://github.com/hbmartin/graphviz2drawio/issues\n")
    stderr.write("and include your diagram and the following information:\n\n")
    stderr.write(DEFAULT_TEXT)
    stderr.write(f"Python: {sys.version}, g2d: {__version__}\n")


def _convert_file(
    to_convert: Path | TextIOWrapper,
    program: str,
    encoding: str,
    outfile: str | None,
) -> None:
    output: str | None = None
    try:
        if isinstance(to_convert, TextIOWrapper):
            output = convert(to_convert, program)
        elif isinstance(to_convert, Path):
            with to_convert.open(encoding=encoding) as contents:
                output = convert(contents.read(), program)
    except UnicodeDecodeError:
        if encoding.lower() != UTF8 and isinstance(to_convert, Path):
            # Attempt to automatically recover for file. Chinese Windows systems in
            # particular often use other encodings e.g. gbk, cp950, cp1252, etc. but
            # the actual dot files are still UTF-8 encoded
            # https://github.com/hbmartin/graphviz2drawio/issues/105
            return _convert_file(to_convert, program, UTF8, outfile)

        _write_stderr_message(str(to_convert))
        raise

    except Exception:
        _write_stderr_message(str(to_convert))
        raise

    if output is None:
        _write_stderr_message(str(to_convert))
        return None

    if outfile is None:
        print(output)
        return None

    out_path = Path(outfile)
    out_path.parent.mkdir(parents=True, exist_ok=True)
    out_path.write_text(output, encoding=encoding)
    stderr.write("Converted file: " + outfile + "\n")
    return None


def main() -> None:
    args = Arguments(__version__).parse_args()  # pytype: disable=not-callable

    in_files: list[TextIOWrapper]
    out_files: list[str | None]

    _validate_args(args)

    if args.stdout and args.outfile is not None:
        sys.stderr.write(f"Writing to {args.outfile} (ignoring stdout)\n")

    if len(args.to_convert) == 1:
        in_files = args.to_convert
        out_files = _determine_single_output(args)
    else:
        in_files = args.to_convert
        out_files = [_gv_filename_to_xml(in_file.name) for in_file in args.to_convert]

    for in_file, out_file in zip(in_files, out_files, strict=True):
        _convert_file(
            to_convert=in_file,
            program=args.program,
            encoding=args.encoding,
            outfile=out_file,
        )


def _determine_single_output(args: Namespace) -> list[str | None]:
    out_files: list[str | None]
    if args.to_convert[0] == sys.stdin:
        out_files = [args.outfile] if args.outfile is not None else [None]
    elif args.outfile is not None:
        out_files = [args.outfile]
    elif args.stdout:
        out_files = [None]
    else:
        out_files = [_gv_filename_to_xml(args.to_convert[0].name)]
    return out_files


def _validate_args(args: Namespace) -> None:
    if len(args.to_convert) > 1 and args.stdout:
        print("Only one file can be converted when using --stdout")
        sys.exit(1)
    if len(args.to_convert) > 1 and args.outfile is not None:
        print("Only one file can be converted when specifying an output file")
        sys.exit(1)
    if len(args.to_convert) == 0 or (
        len(args.to_convert) == 1
        and args.to_convert[0] == sys.stdin
        and sys.stdin.isatty()
    ):
        Arguments(__version__).print_help()  # pytype: disable=not-callable
        sys.exit(1)


if __name__ == "__main__":
    main()



from xml.etree.ElementTree import Element

from ..models import SVG, DotAttr
from ..models.CoordsTranslate import CoordsTranslate
from .CurveFactory import CurveFactory
from .Edge import Edge
from .MxConst import DEFAULT_STROKE_WIDTH
from .Text import Text
from .utils import adjust_color_opacity


class EdgeFactory:
    def __init__(self, coords: CoordsTranslate, *, is_directed: bool) -> None:
        super().__init__()
        self.curve_factory = CurveFactory(coords)
        self.is_directed = is_directed

    def from_svg(self, g: Element, title: str) -> Edge:
        fr, to = title.replace("--", "->").split("->")
        curve = None
        stroke = "#000000"
        stroke_width = DEFAULT_STROKE_WIDTH
        line_style = None
        # pyrefly: ignore  # bad-assignment
        labels: list[Text] = [
            text_from_tag
            for tag in g
            if SVG.is_tag(tag, "text")
            and (text_from_tag := Text.from_svg(tag)) is not None
        ]
        if (path := SVG.get_first(g, "path")) is not None:
            if "d" in path.attrib:
                curve = self.curve_factory.from_svg(path.attrib["d"])
            if "stroke" in path.attrib:
                stroke = path.attrib["stroke"]
                if "stroke-opacity" in path.attrib:
                    stroke = adjust_color_opacity(
                        stroke,
                        float(path.attrib["stroke-opacity"]),
                    )
            if "stroke-width" in path.attrib:
                stroke_width = path.attrib["stroke-width"]
            if "stroke-dasharray" in path.attrib:
                line_style = DotAttr.DASHED
        return Edge(
            sid=g.attrib["id"],
            fr=fr,
            to=to,
            is_directed=self.is_directed,
            curve=curve,
            labels=labels,
            stroke=stroke,
            stroke_width=stroke_width,
            line_style=line_style,
        )



from ..models import DotAttr
from ..models.Rect import Rect
from . import MxConst
from .Curve import Curve
from .GraphObj import GraphObj
from .Styles import Styles
from .Text import Text


class Edge(GraphObj):
    """An edge connecting two nodes in the graph."""

    def __init__(
        self,
        sid: str,
        *,
        fr: str,
        to: str,
        is_directed: bool,
        curve: Curve | None,
        line_style: str | None,
        labels: list[Text],
        stroke: str,
        stroke_width: str,
    ) -> None:
        super().__init__(sid=sid, gid=f"{fr}->{to}")
        self.fr = fr
        self.to = to
        self.curve = curve
        self.line_style = line_style
        self.dir = DotAttr.FORWARD if is_directed else DotAttr.NONE
        self.arrowtail = None
        self.arrowhead = None
        self.labels = labels
        self.stroke = stroke
        self.stroke_width = stroke_width

    def get_edge_style(
        self,
        source_geo: Rect | None,
        target_geo: Rect | None,
    ) -> str:
        dashed = 1 if self.line_style == DotAttr.DASHED else 0
        end_arrow, end_fill = self._get_arrow_shape_and_fill(
            arrow=self.arrowhead,
            active_dirs={DotAttr.FORWARD, DotAttr.BACK, DotAttr.BOTH},
        )
        start_arrow, start_fill = self._get_arrow_shape_and_fill(
            self.arrowtail,
            active_dirs={DotAttr.BOTH},
        )

        if self.curve is not None:
            style = Styles.EDGE.format(
                dashed=dashed,
                end_arrow=end_arrow,
                end_fill=end_fill,
                start_arrow=start_arrow,
                start_fill=start_fill,
                stroke=self.stroke,
                stroke_width=self.stroke_width,
            ) + (MxConst.CURVED if self.curve.is_bezier else MxConst.SHARP)

            if source_geo is not None:
                exit_x, exit_y = source_geo.relative_location_along_perimeter(
                    self.curve.start,
                )
                style += f"exitX={exit_x:.4f};exitY={exit_y:.4f};"
            if target_geo is not None:
                entry_x, entry_y = target_geo.relative_location_along_perimeter(
                    self.curve.end,
                )
                style += f"entryX={entry_x:.4f};entryY={entry_y:.4f};"

            return style

        return Styles.EDGE.format(
            end_arrow=end_arrow,
            dashed=dashed,
            end_fill=end_fill,
            start_arrow=start_arrow,
            start_fill=start_fill,
            stroke=self.stroke,
            stroke_width=self.stroke_width,
        )

    def _get_arrow_shape_and_fill(
        self,
        arrow: str | None,
        active_dirs: set[str],
    ) -> tuple[str, int]:
        shape = MxConst.BLOCK if self.dir in active_dirs else MxConst.NONE
        fill = 1 if self.dir in active_dirs else 0

        if arrow is not None and len(arrow) > 0:
            if arrow == "none":
                shape = MxConst.NONE
                fill = 0
            else:
                if arrow[0] == DotAttr.NO_FILL:
                    fill = 0
                if arrow[1:] == DotAttr.DIAMOND:
                    shape = MxConst.DIAMOND

        return shape, fill

    @property
    def key_for_label(self) -> str:
        return f"{self.gid}-{self.curve}-{self.dir}"

    @property
    def key_for_enrichment(self) -> str:
        cleaned_label = "\\n".join(
            [label.text.replace("\xa0", " ") for label in self.labels],
        )
        return f"{self.gid}-{cleaned_label}"

    def __repr__(self) -> str:
        return (
            f"|{self.fr}->{self.to}|: "
            f"{self.labels}, {self.arrowhead}, {self.dir}, {self.arrowtail};"
        )

    def value_for_labels(self) -> str:
        return "".join(
            (label.get_mx_value() if i == 0 else f"<div>{label.get_mx_value()}</div>")
            for i, label in enumerate(self.labels)
        )



import cmath

# pyrefly: ignore  # import-error
from graphviz2drawio.models.Rect import Rect


def test_closest_point_along_perimeter():
    rect = Rect(x=432.0, y=543.6, width=100.8, height=100.8)
    p1 = rect.closest_point_along_perimeter(544.08, 617.4)
    assert p1 == (532.8, 617.4)


def test_relative_location_along_perimeter():
    rect = Rect(x=432.0, y=543.6, width=100.8, height=100.8)
    p1 = rect.relative_location_along_perimeter(complex(544.08, 617.4))
    assert cmath.isclose(complex(*p1), complex(1.0, 0.73), abs_tol=0.01)



from html import unescape
from xml.etree import ElementTree

COMMENT = "comment"


class CommentedTreeBuilder(ElementTree.TreeBuilder):
    """A TreeBuilder that preserves comments.

    Port forwarding or other stylistic information may be included
    in the <g> tag's <title>, so we prefer to use the preceding
    comment to determine the node's/edge's title instead.
    See `prev_comment` usage in SvgParser.
    """

    def __init__(self):
        super().__init__()
        self.was_root_set = False

    def start(self, tag, attrs):
        elem = super().start(tag, attrs)
        self.was_root_set = True
        return elem

    def comment(self, data):
        rv = super().comment(data)
        if not self.was_root_set:
            return rv
        self.start(COMMENT, {})
        self.data(unescape(data.strip()))
        self.end(COMMENT)
        return rv



_blacklist_attrs = ["fill"]


class GraphObj:

    def __init__(self, sid: str, gid: str) -> None:
        self.sid = sid
        self.gid = gid

    def enrich_from_graph(self, attrs: dict | None) -> None:
        if attrs is None:
            return
        for k, v in attrs.items():
            if v == "" and k in self.__dict__ and self.__dict__[k] is not None:
                continue
            if k in _blacklist_attrs:
                continue
            if k == "shape" and v == "none":
                continue
            self.__setattr__(k, v)



BLOCK = "block"
DIAMOND = "diamond"

CENTER = "center"
START = "start"

DEFAULT_TEXT_SIZE = 10
DEFAULT_FONT_FAMILY = "Helvetica"
DEFAULT_FONT_COLOR = "#000000"
DEFAULT_STROKE_WIDTH = "1"

ARRAY = "Array"
CELL = "mxCell"
GEO = "mxGeometry"
GRAPH = "mxGraphModel"
POINT = "mxPoint"
ROOT = "root"

NONE = "none"
CURVED = "curved=1;"
SHARP = "rounded=0;"

VERTICAL_ALIGN = {
    "t": "top",
    "b": "bottom",
    "c": "middle",
}



from xml.etree.ElementTree import Element

from pygraphviz import AGraph


class GdValueError(ValueError):
    """Base class for exceptions raised during conversion."""

    def __init__(self, message: str) -> None:
        super().__init__(message)


class CouldNotParsePathError(GdValueError):
    """Could not parse path from SVG element."""

    def __init__(self, path: str) -> None:
        super().__init__(
            f"Could not parse path from SVG element: {path}",
        )


class MissingTitleError(GdValueError):
    """Title missing from SVG element."""

    def __init__(self, g: Element) -> None:
        super().__init__(
            f"Title missing from SVG element: {g.tag} with attributes {g.attrib}.",
        )


class MissingTextError(GdValueError):
    """Title missing from SVG element."""

    def __init__(self, g: Element) -> None:
        super().__init__(
            f"Text missing from SVG element: {g.tag} with attributes {g.attrib}.",
        )


class InvalidBezierParameterError(GdValueError):
    """Invalid Bezier parameter, must be 0 <= t <= 1."""

    def __init__(self, t: float) -> None:
        super().__init__(f"Invalid Bezier parameter (t={t}), must be 0 <= t <= 1.")


class MissingIdentifiersError(GdValueError):
    """Missing identifiers for a geometry."""

    def __init__(self, sid: str | None, gid: str | None) -> None:
        super().__init__(
            f"Missing identifiers for a geometry: sid(id): {sid}, gid(title): {gid}.",
        )


class UnableToParseGraphError(GdValueError):
    """Graph was unexpectedly None."""

    def __init__(self, graph: AGraph) -> None:
        super().__init__(
            f"Graph.draw() returned by pygraphviz was unexpectedly None: {graph}",
        )



# pyrefly: ignore  # import-error
from graphviz2drawio.mx.Curve import _line


def test_line():
    line = _line(complex(-5, 10), complex(-3, 4))
    assert line(0) == -5
    assert line(1) == -8
    assert line(2) == -11


def test_vertical_line_returns_none():
    line = _line(complex(1, 10), complex(1, 4))
    assert line is None



class CoordsTranslate:
    def __init__(self, x: float, y: float) -> None:
        self.x = x
        self.y = y

    def complex_translate(self, cnum: complex) -> complex:
        return complex(cnum.real + self.x, cnum.imag + self.y)

    def translate(self, x: float | str, y: float | str) -> tuple[float, float]:
        return float(x) + self.x, float(y) + self.y

    @staticmethod
    def from_svg_transform(transform: str) -> "CoordsTranslate":
        x, y = transform.split("translate(")[1].split(")")[0].split(" ")
        return CoordsTranslate(x=float(x), y=float(y))



from typing import Optional
from xml.etree.ElementTree import Element

from . import MxConst
from .Styles import Styles


class Text:
    def __init__(
        self,
        text: str,
        family: str,
        size: float,
        color: str | None,
        *,
        bold: bool,
        italic: bool,
    ) -> None:
        self.family = family
        self.size = size
        self.text = text
        self.color = color
        self.bold = bold
        self.italic = italic

    def get_mx_value(self) -> str:
        open_tags = ""
        close_tags = ""
        if self.bold:
            open_tags = "<b>"
            close_tags = "</b>"
        if self.italic:
            open_tags = "<i>" + open_tags
            close_tags += "</i>"
        return Styles.TEXT_VALUE.format(
            text=self.text,
            size=self.size or "14",
            family=self.family or MxConst.DEFAULT_FONT_FAMILY,
            color=self.color or MxConst.DEFAULT_FONT_COLOR,
            open_tags=open_tags,
            close_tags=close_tags,
        )

    @staticmethod
    def from_svg(t: Element) -> Optional["Text"]:
        text = t.text
        if text is None or text.strip() == "":
            return None
        return Text(
            text=text.replace("<", "&lt;").replace(">", "&gt;"),
            family=t.attrib.get("font-family", MxConst.DEFAULT_FONT_FAMILY),
            size=float(t.attrib.get("font-size", MxConst.DEFAULT_TEXT_SIZE)),
            color=t.attrib.get("fill", MxConst.DEFAULT_FONT_COLOR),
            bold=t.get("font-weight", None) == "bold",
            italic=t.get("font-style", None) == "italic",
        )

    def __repr__(self) -> str:
        return self.text



from enum import Enum

from . import Shape


# Make this subclass StrEnum when dropping Py 3.10 support
class Styles(Enum):
    NODE = (
        "verticalAlign={vertical_align};html=1;rounded=0;labelBackgroundColor=none;strokeColor={stroke};"
        "fillColor={fill};strokeWidth={stroke_width};dashed={dashed};"
    )
    EDGE = (
        "html=1;endArrow={end_arrow};dashed={dashed};endFill={end_fill};startArrow={start_arrow};"
        "startFill={start_fill};fillColor={stroke};strokeColor={stroke};strokeWidth={stroke_width};"
    )
    EDGE_LABEL = (
        "edgeLabel;html=1;align=center;verticalAlign=bottom;resizable=0;points=[];"
    )
    EDGE_INVIS = (
        "rounded=1;html=1;exitX={exit_x:.3g};exitY={exit_y:.3g};jettySize=auto;curved={curved};"
        "endArrow={end_arrow};dashed={dashed};endFill={end_fill};"
    )

    TEXT_VALUE = (
        "{open_tags}<font style='font-size: {size}px;' face='{family}' color='{color}'>"
        "{text}</font>{close_tags}"
    )

    ELLIPSE = "ellipse;" + NODE
    CIRCLE = "ellipse;aspect=fixed;" + NODE
    HEXAGON = "shape=hexagon;perimeter=hexagonPerimeter2;" + NODE
    EGG = "shape=mxgraph.flowchart.display;direction=south;" + NODE
    TRIANGLE = "triangle;direction=north;" + NODE
    DIAMOND = "rhombus;" + NODE
    TRAPEZOID = "shape=trapezoid;perimeter=trapezoidPerimeter;" + NODE
    PARALLELOGRAM = "shape=parallelogram;perimeter=parallelogramPerimeter;" + NODE
    HOUSE = "shape=offPageConnector;direction=west;" + NODE
    PENTAGON = "shape=mxgraph.basic.pentagon;" + NODE
    OCTAGON = "shape=mxgraph.basic.octagon2;align=center;dx=15;" + NODE
    DOUBLE_CIRCLE = "ellipse;shape=doubleEllipse;aspect=fixed;" + NODE
    DOUBLE_OCTAGON = (
        "shape=image;html=1;verticalLabelPosition=middle;imageAspect=0;aspect=fixed;"
        "image=https://cdn4.iconfinder.com/data/icons/feather/24/octagon-128.png;labelPosition=center;align=center;"
        + NODE
    )
    INV_TRIANGLE = "triangle;direction=south;" + NODE
    INV_TRAPEZOID = (
        "shape=trapezoid;perimeter=trapezoidPerimeter;direction=west;" + NODE
    )
    INV_HOUSE = "shape=offPageConnector;direction=east;" + NODE
    SQUARE = "aspect=fixed;" + NODE
    STAR = (
        "shape=mxgraph.basic.star;labelPosition=center;align=center;verticalLabelPosition=middle;"
        + NODE
    )
    UNDERLINE = (
        "line;strokeWidth=2;verticalAlign=bottom;labelPosition=center;"
        "verticalLabelPosition=top;align=center;"
    )
    CYLINDER = "shape=cylinder;boundedLbl=1;backgroundOutline=1;" + NODE
    NOTE = "shape=note;backgroundOutline=1;" + NODE
    TAB = "shape=folder;tabWidth=40;tabHeight=14;tabPosition=left;" + NODE
    FOLDER = (
        "shape=mxgraph.office.concepts.folder;outlineConnect=0;align=center;verticalLabelPosition=middle;"
        "labelPosition=center;shadow=0;dashed=0;" + NODE
    )
    CUBE = "shape=cube;boundedLbl=1;backgroundOutline=1;" + NODE
    COMPONENT = "shape=component;align=center;spacingLeft=36;" + NODE
    RPROMOTER = (
        "shape=mxgraph.arrows2.bendArrow;dy=15;dx=38;notch=0;arrowHead=55;rounded=0;shadow=0;dashed=0;align=center;"
        + NODE
    )
    LPROMOTER = "flipH=1;" + RPROMOTER
    CDS = (
        "shape=mxgraph.arrows2.arrow;dy=0;dx=10;notch=0;shadow=0;dashed=0;align=center;"
        + NODE
    )
    RARROW = (
        "shape=mxgraph.arrows2.arrow;dy=0.6;dx=40;align=center;labelPosition=center;notch=0;strokeWidth=2;"
        "verticalLabelPosition=middle;" + NODE
    )
    LARROW = "flipH=1;" + RARROW
    IMAGE = (
        "shape=image;verticalLabelPosition=top;labelBackgroundColor=default;verticalAlign=top;aspect=fixed;"
        "imageAspect=0;image={image};" + NODE
    )

    @staticmethod
    def get_for_shape(dot_shape: str | None) -> "Styles":
        if dot_shape is None:
            return Styles.ELLIPSE
        if dot_shape.startswith("M") and dot_shape[1:] in _shape_to_style:
            return _shape_to_style[dot_shape[1:]]
        if dot_shape in _shape_to_style:
            return _shape_to_style[dot_shape]
        return Styles.NODE

    def format(self, **values) -> str:  # noqa: ANN003
        return self.value.format(**values)


_shape_to_style: dict[str, "Styles"] = {
    Shape.BOX: Styles.NODE,
    Shape.BOX_3D: Styles.CUBE,
    Shape.CDS: Styles.CDS,
    Shape.CIRCLE: Styles.CIRCLE,
    Shape.COMPONENT: Styles.COMPONENT,
    Shape.CYLINDER: Styles.CYLINDER,
    Shape.DIAMOND: Styles.DIAMOND,
    Shape.DOUBLE_CIRCLE: Styles.DOUBLE_CIRCLE,
    Shape.DOUBLE_OCTAGON: Styles.DOUBLE_OCTAGON,
    Shape.EGG: Styles.EGG,
    Shape.ELLIPSE: Styles.ELLIPSE,
    Shape.FOLDER: Styles.FOLDER,
    Shape.HEXAGON: Styles.HEXAGON,
    Shape.HOUSE: Styles.HOUSE,
    Shape.IMAGE: Styles.IMAGE,
    Shape.INV_HOUSE: Styles.INV_HOUSE,
    Shape.INV_TRAPEZIUM: Styles.INV_TRAPEZOID,
    Shape.INV_TRIANGLE: Styles.INV_TRIANGLE,
    Shape.LARROW: Styles.LARROW,
    Shape.LPROMOTER: Styles.LPROMOTER,
    Shape.NOTE: Styles.NOTE,
    Shape.OCTAGON: Styles.OCTAGON,
    Shape.OVAL: Styles.ELLIPSE,
    Shape.PARALLELOGRAM: Styles.PARALLELOGRAM,
    Shape.PENTAGON: Styles.PENTAGON,
    Shape.PLAIN: Styles.NODE,
    Shape.POLYGON: Styles.HEXAGON,
    Shape.PROMOTER: Styles.RPROMOTER,
    Shape.RARROW: Styles.RARROW,
    Shape.RECT: Styles.NODE,
    Shape.RECTANGLE: Styles.NODE,
    Shape.RPROMOTER: Styles.RPROMOTER,
    Shape.SQUARE: Styles.SQUARE,
    Shape.STAR: Styles.STAR,
    Shape.TAB: Styles.TAB,
    Shape.TRAPEZIUM: Styles.TRAPEZOID,
    Shape.TRIANGLE: Styles.TRIANGLE,
    Shape.UNDERLINE: Styles.UNDERLINE,
}



import html
import re
from pathlib import Path
from xml.etree import ElementTree

# pyrefly: ignore  # missing-module-attribute
from graphviz2drawio import graphviz2drawio

num_cells_offset = 2


def check_xml_top(root):
    graph = root.findall(".")
    assert len(graph) == 1
    graph = graph[0]
    assert graph.tag == "mxGraphModel"
    return graph.findall("./root/*")


def check_style(e, check) -> None:
    assert check in e.attrib["style"]


def check_value(e, check) -> None:
    value = e.attrib["value"]
    match = re.match(f"<font.*>{check}</font>", html.unescape(value))
    assert match, f"no match found for {value}"


def check_edge(edge, source, target) -> None:
    assert edge.attrib.get("edge")
    assert edge.attrib["source"] == source.attrib["id"]
    assert edge.attrib["target"] == target.attrib["id"]


def check_edge_dir(e, dx, dy) -> None:
    style = [e.split("=") for e in e.attrib["style"].split(";")][:-1]
    style = {e[0]: e[1] for e in style if e}
    x2 = float(style["exitX"])
    x1 = float(style["entryX"])
    y2 = float(style["exitY"])
    y1 = float(style["entryY"])

    assert (x2 - x1) == dx
    assert (y2 - y1) == dy


def test_hello() -> None:
    file = "test/directed/hello.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    elements = check_xml_top(root)

    hello = elements[2]
    check_style(hello, "ellipse")
    check_value(hello, "Hello")

    world = elements[3]
    check_style(world, "ellipse")
    check_value(world, "World")
    edge = elements[4]
    check_edge(edge, hello, world)


def test_hello_after_reading_content_with_comment() -> None:
    content_with_comment = Path("test/directed/hello.gv.txt").read_text()
    xml = graphviz2drawio.convert(content_with_comment)

    root = ElementTree.fromstring(xml)
    elements = check_xml_top(root)

    hello = elements[2]
    check_style(hello, "ellipse")
    check_value(hello, "Hello")

    world = elements[3]
    check_style(world, "ellipse")
    check_value(world, "World")
    edge = elements[4]
    check_edge(edge, hello, world)


def test_hello_rect() -> None:
    file = "test/directed/hello_rect.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    elements = check_xml_top(root)

    hello = elements[2]
    check_value(hello, "Hello")
    assert "ellipse" not in hello.attrib["style"]

    world = elements[3]
    check_value(world, "World")
    assert "ellipse" not in world.attrib["style"]


def test_port() -> None:
    file = "test/directed/port.gv.txt"
    xml = graphviz2drawio.convert(file)
    root = ElementTree.fromstring(xml)
    elements = check_xml_top(root)
    check_edge(elements[6], elements[2], elements[3])


def test_polylines() -> None:
    file = "test/undirected/polylines.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    check_xml_top(root)
    assert "curved" not in xml
    assert "rounded=0" in xml


def test_polylines_curved() -> None:
    file = "test/undirected/polylines_curved.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    check_xml_top(root)

    assert "curved=1" in xml


def test_cluster() -> None:
    file = "test/directed/cluster.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)

    elements = check_xml_top(root)
    contains_cluster = False
    for el in elements:
        if "process" in el.attrib.get("value", ""):
            contains_cluster = True
    assert contains_cluster


def test_convnet() -> None:
    file = "test/directed/convnet.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    elements = check_xml_top(root)

    assert (
        elements[-1].attrib["value"]
        == "<font style='font-size: 14.0px;' face='Times,serif' color='#000000'>$$l_t$$</font>"
    )
    assert "doubleEllipse" in xml
    assert "steelblue1" not in xml


def test_multilabel() -> None:
    file = "test/directed/multilabel.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    elements = check_xml_top(root)

    assert (
        elements[-1].attrib["value"]
        == "<font style='font-size: 14.0px;' face='Times,serif' color='#000000'>c</font><div><font style='font-size: 14.0px;' face='Times,serif' color='#000000'>b</font></div><div><font style='font-size: 14.0px;' face='Times,serif' color='#000000'>a</font></div>"
    )


def test_datastruct() -> None:
    file = "test/directed/datastruct.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    elements = check_xml_top(root)

    assert elements[-1].attrib["source"] == "node12"
    assert elements[-1].attrib["target"] == "node2"


def test_compound() -> None:
    file = "test/directed/compound.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    elements = check_xml_top(root)
    assert elements[2].attrib["id"] == "clust1"
    assert elements[3].attrib["id"] == "clust2"


def test_subgraph_and_colors():
    file = "test/directed/subgraph_multiple.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    check_xml_top(root)
    assert "fillColor=none" in xml
    assert "strokeColor=black" in xml
    assert "clust1" in xml
    assert "clust2" in xml


def test_title_with_colon():
    file = "test/directed/bazel.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    check_xml_top(root)
    assert "//absl/random:random&lt;/font&gt;" in xml


def test_invisible():
    file = "test/directed/invisible.gv.txt"
    xml = graphviz2drawio.convert(file)

    root = ElementTree.fromstring(xml)
    check_xml_top(root)
    assert ';exitX=1.0000;exitY=0.5839;"' in xml



BOX = "box"
BOX_3D = "box3d"
CDS = "cds"
CIRCLE = "circle"
COMPONENT = "component"
CYLINDER = "cylinder"
DIAMOND = "diamond"
DOUBLE_CIRCLE = "doublecircle"
DOUBLE_OCTAGON = "doubleoctagon"
EGG = "egg"
ELLIPSE = "ellipse"
FOLDER = "folder"
HEXAGON = "hexagon"
HOUSE = "house"
IMAGE = "image"
INV_HOUSE = "invhouse"
INV_TRAPEZIUM = "invtrapezium"
INV_TRIANGLE = "invtriangle"
LARROW = "larrow"
LPROMOTER = "lpromoter"
NOTE = "note"
OCTAGON = "octagon"
OVAL = "oval"
PARALLELOGRAM = "parallelogram"
PENTAGON = "pentagon"
PLAIN = "plain"
POLYGON = "polygon"
PROMOTER = "promoter"
RARROW = "rarrow"
RECT = "rect"
RECTANGLE = "rectangle"
RPROMOTER = "rpromoter"
SQUARE = "square"
STAR = "star"
TAB = "tab"
TRAPEZIUM = "trapezium"
TRIANGLE = "triangle"
UNDERLINE = "underline"



import cmath
from typing import Any

from svg.path import CubicBezier, Path, QuadraticBezier, parse_path

from ..models.CoordsTranslate import CoordsTranslate
from .bezier import approximate_cubic_bezier_as_quadratic, subdivide_inflections
from .Curve import Curve

CLOSE_POINT_TOLERANCE = 0.1


class CurveFactory:
    def __init__(self, coords: CoordsTranslate) -> None:
        super().__init__()
        self.coords = coords

    def from_svg(self, svg_path: str) -> Curve:
        path: Path = parse_path(svg_path)
        points: list[complex] = []
        is_bezier = not all(map(Curve.is_linear, filter(_is_cubic, path)))

        for segment in path:
            if isinstance(segment, QuadraticBezier):
                points.append(self.coords.complex_translate(segment.control))
            elif isinstance(segment, CubicBezier):
                if Curve.is_linear(segment):
                    points.append(self.coords.complex_translate(segment.start))
                else:
                    split_cubes = subdivide_inflections(
                        segment.start,
                        segment.control1,
                        segment.control2,
                        segment.end,
                    )
                    split_controls = [
                        self.coords.complex_translate(
                            approximate_cubic_bezier_as_quadratic(*cube)[1],
                        )
                        for cube in split_cubes
                        if cube
                    ]
                    points.extend(split_controls)

        # pyrefly: ignore  # missing-attribute
        start: complex = self.coords.complex_translate(path[0].start)
        # pyrefly: ignore  # missing-attribute
        end: complex = self.coords.complex_translate(path[-1].end)

        if len(points) > 0 and cmath.isclose(
            start,
            points[0],
            rel_tol=CLOSE_POINT_TOLERANCE,
        ):
            points = points[1:]

        return Curve(start=start, end=end, is_bezier=is_bezier, points=points)


def _is_cubic(p: Any) -> bool:  # noqa: ANN401
    return isinstance(p, CubicBezier)



from os import path

# pyrefly: ignore  # import-error
from setuptools import find_packages, setup

here = path.abspath(path.dirname(__file__))

with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

with open(path.join(here, "graphviz2drawio", "version.py")) as fp:
    exec(fp.read())

setup(
    name="graphviz2drawio",
    # pyrefly: ignore  # unknown-name
    version=__version__,  # noqa: F821
    description="Convert graphviz (dot) files to draw.io / lucid (mxGraph) format. "
    "Beautiful and editable graphs in your favorite editor.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/hbmartin/graphviz2drawio/",
    author="Harold Martin",
    author_email="harold.martin@gmail.com",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Information Technology",
        "Intended Audience :: Science/Research",
        "Topic :: Multimedia :: Graphics",
        "Topic :: Multimedia :: Graphics :: Graphics Conversion",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Scientific/Engineering :: Visualization",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: 3.14",
    ],
    license="GPLv3",
    keywords="graphviz graph agraph dot convert conversion diagrams drawio mxgraph maxgraph Lucidchart",
    packages=find_packages(exclude=["doc", "test"]),
    python_requires=">=3.10",
    install_requires=["puremagic", "pygraphviz", "svg.path"],
    tests_require=["diagrams", "pytest"],
    entry_points={"console_scripts": ["graphviz2drawio=graphviz2drawio.__main__:main"]},
    project_urls={
        "Bug Reports": "https://github.com/hbmartin/graphviz2drawio/issues",
        "Source": "https://github.com/hbmartin/graphviz2drawio/",
    },
    include_package_data=True,
    package_data={"graphviz2drawio": ["py.typed"]},
)



def adjust_color_opacity(hex_color: str, opacity: float) -> str:
    """Adjust the opacity of a hex color as if it were on a white background."""
    # Remove the '#' if present
    hex_color = hex_color.lstrip("#")

    # Convert hex to RGB
    try:
        r, g, b = tuple(int(hex_color[i : i + 2], 16) for i in (0, 2, 4))
    except ValueError:
        return hex_color

    # Apply opacity over white background
    r = int(r * opacity + 255 * (1 - opacity))
    g = int(g * opacity + 255 * (1 - opacity))
    b = int(b * opacity + 255 * (1 - opacity))

    # Convert back to hex
    return f"#{r:02x}{g:02x}{b:02x}"



from svg import path as svg_path

from ..models.CoordsTranslate import CoordsTranslate
from ..models.Errors import CouldNotParsePathError
from ..models.Rect import Rect


def rect_from_svg_path(coords: CoordsTranslate, path_d: str) -> Rect:
    parsed_path: svg_path.Path = svg_path.parse_path(path_d)
    if len(parsed_path) == 0 or not isinstance(parsed_path[0], svg_path.Move):
        raise CouldNotParsePathError(path_d)
    start: svg_path.Move = parsed_path.pop(0)
    min_x = start.start.real
    min_y = start.start.imag
    max_x = start.start.real
    max_y = start.start.imag

    # Note that this loop may not be accurate since it does not calculate
    # the Bezier curve based on the control points
    for e in parsed_path:
        min_x = min(min_x, e.start.real, e.end.real)
        min_y = min(min_y, e.start.imag, e.end.imag)
        max_x = max(max_x, e.start.real, e.end.real)
        max_y = max(max_y, e.start.imag, e.end.imag)

    (x, y) = coords.translate(min_x, min_y)

    return Rect(x=x, y=y, width=max_x - min_x, height=max_y - min_y)


def rect_from_svg_points(coords: CoordsTranslate, svg: str) -> Rect:
    """Don't use this function.

    It will be replaced by accurately returning the shape of the SVG.
    """
    points: list[tuple[float, float]] = [
        coords.translate(*p.split(",")) for p in svg.split(" ")
    ]
    min_x, min_y = points[0]
    width = 0.0
    height = 0.0
    for p in points:
        min_x = min(p[0], min_x)
        min_y = min(p[1], min_y)
    for p in points:
        test_width = p[0] - min_x
        test_height = p[1] - min_y
        width = max(test_width, width)
        height = max(test_height, height)
    return Rect(x=min_x, y=min_y, width=width, height=height)


def rect_from_image(coords: CoordsTranslate, attrib: dict[str, str]) -> Rect:
    x, y = coords.translate(attrib["x"], attrib["y"])

    return Rect(
        width=float(attrib["width"].strip("px")),
        height=float(attrib["height"].strip("px")),
        x=x,
        y=y,
        image=attrib["{http://www.w3.org/1999/xlink}href"],
    )


def rect_from_ellipse_svg(coords: CoordsTranslate, attrib: dict[str, str]) -> Rect:
    cx = float(attrib["cx"])
    cy = float(attrib["cy"])
    rx = float(attrib["rx"])
    ry = float(attrib["ry"])
    x, y = coords.translate(cx, cy)
    return Rect(x=x - rx, y=y - ry, width=rx * 2, height=ry * 2)



import re
from xml.etree.ElementTree import Element

from ..models import SVG
from ..models.CoordsTranslate import CoordsTranslate
from ..models.Errors import MissingIdentifiersError
from . import MxConst, Shape
from .MxConst import DEFAULT_STROKE_WIDTH
from .Node import Gradient, Node
from .RectFactory import (
    rect_from_ellipse_svg,
    rect_from_image,
    rect_from_svg_path,
    rect_from_svg_points,
)
from .Text import Text
from .utils import adjust_color_opacity


class NodeFactory:
    def __init__(self, coords: CoordsTranslate) -> None:
        super().__init__()
        self.coords = coords

    def from_svg(
        self,
        g: Element,
        labelloc: str,
        gradients: dict[str, Gradient],
    ) -> Node:
        sid = g.attrib["id"]
        gid = SVG.get_title(g)
        rect = None
        fill: str | Gradient = MxConst.NONE
        stroke = MxConst.NONE
        stroke_width = DEFAULT_STROKE_WIDTH
        dashed = False

        if sid is None or gid is None:
            raise MissingIdentifiersError(sid, gid)

        if (inner_g := SVG.get_first(g, "g")) is not None:
            if (inner_a := SVG.get_first(inner_g, "a")) is not None:
                g = inner_a

        if (polygon := SVG.get_first(g, "polygon")) is not None:
            rect = rect_from_svg_points(self.coords, polygon.attrib["points"])
            shape = Shape.RECT
            fill = self._extract_fill(polygon, gradients)
            stroke = self._extract_stroke(polygon)
            stroke_width = polygon.attrib.get("stroke-width", DEFAULT_STROKE_WIDTH)
            if "stroke-dasharray" in polygon.attrib:
                dashed = True
        elif (image := SVG.get_first(g, "image")) is not None:
            rect = rect_from_image(self.coords, image.attrib)
            shape = Shape.IMAGE
        elif (ellipse := SVG.get_first(g, "ellipse")) is not None:
            rect = rect_from_ellipse_svg(
                coords=self.coords,
                attrib=ellipse.attrib,  # pytype: disable=attribute-error
            )
            shape = (
                Shape.ELLIPSE
                if len(SVG.findall(g, "ellipse")) == 1
                else Shape.DOUBLE_CIRCLE
            )
            fill = self._extract_fill(ellipse, gradients)
            stroke = self._extract_stroke(ellipse)
            stroke_width = ellipse.attrib.get("stroke-width", DEFAULT_STROKE_WIDTH)
            if "stroke-dasharray" in ellipse.attrib:
                dashed = True
        elif (path := SVG.get_first(g, "path")) is not None:
            rect = rect_from_svg_path(self.coords, path.attrib["d"])
            shape = Shape.RECT
            fill = self._extract_fill(path, gradients)
            stroke = self._extract_stroke(path)
        else:
            shape = Shape.ELLIPSE

        texts, text_offset = self._extract_texts(g)

        return Node(
            sid=sid,
            gid=gid,
            rect=rect,
            texts=texts,
            fill=fill,
            stroke=stroke,
            shape=shape,
            labelloc=labelloc,
            stroke_width=stroke_width,
            text_offset=text_offset,
            dashed=dashed,
        )

    _fill_url_re = re.compile(r"url\(#([^)]+)\)")

    @staticmethod
    def _extract_fill(g: Element, gradients: dict[str, Gradient]) -> str | Gradient:
        fill = g.attrib.get("fill", MxConst.NONE)
        if fill.startswith("url"):
            match = NodeFactory._fill_url_re.search(fill)
            if match is not None:
                return gradients[match.group(1)]
        if "fill-opacity" in g.attrib and fill != MxConst.NONE:
            fill = adjust_color_opacity(fill, float(g.attrib["fill-opacity"]))
        return fill

    @staticmethod
    def _extract_stroke(g: Element) -> str:
        stroke = g.attrib.get("stroke", MxConst.NONE)
        if "stroke-opacity" in g.attrib and stroke != MxConst.NONE:
            stroke = adjust_color_opacity(stroke, float(g.attrib["stroke-opacity"]))
        return stroke

    def _extract_texts(self, g: Element) -> tuple[list[Text], complex | None]:
        texts = []
        current_text: Text | None = None
        offset: complex | None = None
        # pyrefly: ignore  # unknown
        for t in g:
            if SVG.is_tag(t, "text"):
                if current_text is None:
                    current_text = Text.from_svg(t)
                else:
                    current_text.text += f"<br/>{t.text}"
                if offset is None and current_text is not None:
                    x = None
                    y = None
                    try:
                        x = float(t.attrib.get("x", "x"))
                        y = float(t.attrib.get("y", "y")) - current_text.size
                    except ValueError:
                        pass
                    if x is not None and y is not None:
                        offset = self.coords.complex_translate(complex(x, y))
            elif current_text is not None:
                texts.append(current_text)
                current_text = None
        if current_text is not None:
            texts.append(current_text)
        return texts, offset



from typing import TypeAlias

from ..models.Rect import Rect
from .GraphObj import GraphObj
from .MxConst import VERTICAL_ALIGN
from .Styles import Styles
from .Text import Text

Gradient: TypeAlias = tuple[str, str | None, str]


class Node(GraphObj):
    def __init__(
        self,
        sid: str,
        gid: str,
        rect: Rect | None,
        texts: list[Text],
        fill: str | Gradient,
        stroke: str,
        shape: str,
        labelloc: str,
        stroke_width: str,
        text_offset: complex | None,
        *,
        dashed: bool,
    ) -> None:
        super().__init__(sid, gid)
        self.rect = rect
        self.texts = texts
        self.fill = fill
        self.stroke = stroke
        self.label = None
        self.shape = shape
        self.labelloc = labelloc
        self.stroke_width = stroke_width
        self.text_offset = text_offset
        self.dashed = dashed

    def texts_to_mx_value(self) -> str:
        value = ""
        last_text = len(self.texts) - 1
        for i, t in enumerate(self.texts):
            value += f"<div>{t.get_mx_value()}</div>" if i != 0 else t.get_mx_value()
            if i != last_text:
                value += "<hr size='1'/>"

        return value

    def get_node_style(self) -> str:
        style_for_shape = Styles.get_for_shape(self.shape)
        dashed = 1 if self.dashed else 0
        additional_styling = ""

        attributes = {
            "stroke": self.stroke,
            "stroke_width": self.stroke_width,
            "dashed": dashed,
        }
        if isinstance(self.fill, str):
            attributes["fill"] = self.fill
        elif type(self.fill) is tuple:
            attributes["fill"] = self.fill[0]
            additional_styling += (
                f"gradientColor={self.fill[1]};gradientDirection={self.fill[2]};"
            )

        if (rect := self.rect) is not None and (image_path := rect.image) is not None:
            from .image import image_data_for_path

            attributes["image"] = image_data_for_path(image_path)

        attributes["vertical_align"] = VERTICAL_ALIGN.get(self.labelloc, "middle")

        return style_for_shape.format(**attributes) + additional_styling

    def __repr__(self) -> str:
        return (
            f"Node({self.sid}, {self.gid}, {self.fill}, {self.stroke}, {self.shape}, "
            f"{self.rect})"
        )
